@@ -2,7 +2,75 @@
 #include "ruby/encoding.h"
 
 static VALUE rb_cPathname;
-static ID id_at_path, id_to_path;
+static ID id_ENOTDIR;
+static ID id_at_path;
+static ID id_atime;
+static ID id_base;
+static ID id_basename;
+static ID id_binread;
+static ID id_binwrite;
+static ID id_birthtime;
+static ID id_blockdev_p;
+static ID id_chardev_p;
+static ID id_chmod;
+static ID id_chown;
+static ID id_ctime;
+static ID id_directory_p;
+static ID id_dirname;
+static ID id_empty_p;
+static ID id_entries;
+static ID id_executable_p;
+static ID id_executable_real_p;
+static ID id_exist_p;
+static ID id_expand_path;
+static ID id_extname;
+static ID id_file_p;
+static ID id_fnmatch;
+static ID id_foreach;
+static ID id_ftype;
+static ID id_getwd;
+static ID id_glob;
+static ID id_grpowned_p;
+static ID id_lchmod;
+static ID id_lchown;
+static ID id_link;
+static ID id_lstat;
+static ID id_mkdir;
+static ID id_mtime;
+static ID id_open;
+static ID id_owned_p;
+static ID id_pipe_p;
+static ID id_read;
+static ID id_readable_p;
+static ID id_readable_real_p;
+static ID id_readlines;
+static ID id_readlink;
+static ID id_realdirpath;
+static ID id_realpath;
+static ID id_rename;
+static ID id_rmdir;
+static ID id_setgid_p;
+static ID id_setuid_p;
+static ID id_size;
+static ID id_size_p;
+static ID id_socket_p;
+static ID id_split;
+static ID id_stat;
+static ID id_sticky_p;
+static ID id_sub;
+static ID id_symlink;
+static ID id_symlink_p;
+static ID id_sysopen;
+static ID id_to_path;
+static ID id_truncate;
+static ID id_unlink;
+static ID id_utime;
+static ID id_world_readable_p;
+static ID id_world_writable_p;
+static ID id_writable_p;
+static ID id_writable_real_p;
+static ID id_write;
+static ID id_zero_p;
 
 static VALUE
 get_strpath(VALUE obj)
@@ -155,11 +223,15 @@ path_cmp(VALUE self, VALUE other)
     return INT2FIX(0);
 }
 
+#ifndef ST2FIX
+#define ST2FIX(h) LONG2FIX((long)(h))
+#endif
+
 /* :nodoc: */
 static VALUE
 path_hash(VALUE self)
 {
-    return INT2FIX(rb_str_hash(get_strpath(self)));
+    return ST2FIX(rb_str_hash(get_strpath(self)));
 }
 
 /*
@@ -199,10 +271,10 @@ path_sub(int argc, VALUE *argv, VALUE self)
     VALUE str = get_strpath(self);
 
     if (rb_block_given_p()) {
-        str = rb_block_call(str, rb_intern("sub"), argc, argv, 0, 0);
+        str = rb_block_call(str, id_sub, argc, argv, 0, 0);
     }
     else {
-        str = rb_funcallv(str, rb_intern("sub"), argc, argv);
+        str = rb_funcallv(str, id_sub, argc, argv);
     }
     return rb_class_new_instance(1, &str, rb_obj_class(self));
 }
@@ -257,7 +329,7 @@ path_realpath(int argc, VALUE *argv, VALUE self)
 {
     VALUE basedir, str;
     rb_scan_args(argc, argv, "01", &basedir);
-    str = rb_funcall(rb_cFile, rb_intern("realpath"), 2, get_strpath(self), basedir);
+    str = rb_funcall(rb_cFile, id_realpath, 2, get_strpath(self), basedir);
     return rb_class_new_instance(1, &str, rb_obj_class(self));
 }
 
@@ -273,7 +345,7 @@ path_realdirpath(int argc, VALUE *argv, VALUE self)
 {
     VALUE basedir, str;
     rb_scan_args(argc, argv, "01", &basedir);
-    str = rb_funcall(rb_cFile, rb_intern("realdirpath"), 2, get_strpath(self), basedir);
+    str = rb_funcall(rb_cFile, id_realdirpath, 2, get_strpath(self), basedir);
     return rb_class_new_instance(1, &str, rb_obj_class(self));
 }
 
@@ -296,10 +368,10 @@ path_each_line(int argc, VALUE *argv, VALUE self)
     args[0] = get_strpath(self);
     n = rb_scan_args(argc, argv, "03", &args[1], &args[2], &args[3]);
     if (rb_block_given_p()) {
-        return rb_block_call(rb_cIO, rb_intern("foreach"), 1+n, args, 0, 0);
+        return rb_block_call(rb_cFile, id_foreach, 1+n, args, 0, 0);
     }
     else {
-        return rb_funcallv(rb_cIO, rb_intern("foreach"), 1+n, args);
+        return rb_funcallv(rb_cFile, id_foreach, 1+n, args);
     }
 }
 
@@ -310,7 +382,7 @@ path_each_line(int argc, VALUE *argv, VALUE self)
  *
  * Returns all data from the file, or the first +N+ bytes if specified.
  *
- * See IO.read.
+ * See File.read.
  *
  */
 static VALUE
@@ -321,7 +393,7 @@ path_read(int argc, VALUE *argv, VALUE self)
 
     args[0] = get_strpath(self);
     n = rb_scan_args(argc, argv, "03", &args[1], &args[2], &args[3]);
-    return rb_funcallv(rb_cIO, rb_intern("read"), 1+n, args);
+    return rb_funcallv_kw(rb_cFile, id_read, 1+n, args, RB_PASS_CALLED_KEYWORDS);
 }
 
 /*
@@ -330,7 +402,7 @@ path_read(int argc, VALUE *argv, VALUE self)
  *
  * Returns all the bytes from the file, or the first +N+ if specified.
  *
- * See IO.binread.
+ * See File.binread.
  *
  */
 static VALUE
@@ -341,7 +413,7 @@ path_binread(int argc, VALUE *argv, VALUE self)
 
     args[0] = get_strpath(self);
     n = rb_scan_args(argc, argv, "02", &args[1], &args[2]);
-    return rb_funcallv(rb_cIO, rb_intern("binread"), 1+n, args);
+    return rb_funcallv(rb_cFile, id_binread, 1+n, args);
 }
 
 /*
@@ -351,7 +423,7 @@ path_binread(int argc, VALUE *argv, VALUE self)
  *
  * Writes +contents+ to the file.
  *
- * See IO.write.
+ * See File.write.
  *
  */
 static VALUE
@@ -362,7 +434,7 @@ path_write(int argc, VALUE *argv, VALUE self)
 
     args[0] = get_strpath(self);
     n = rb_scan_args(argc, argv, "03", &args[1], &args[2], &args[3]);
-    return rb_funcallv(rb_cIO, rb_intern("write"), 1+n, args);
+    return rb_funcallv_kw(rb_cFile, id_write, 1+n, args, RB_PASS_CALLED_KEYWORDS);
 }
 
 /*
@@ -372,7 +444,7 @@ path_write(int argc, VALUE *argv, VALUE self)
  *
  * Writes +contents+ to the file, opening it in binary mode.
  *
- * See IO.binwrite.
+ * See File.binwrite.
  *
  */
 static VALUE
@@ -383,7 +455,7 @@ path_binwrite(int argc, VALUE *argv, VALUE self)
 
     args[0] = get_strpath(self);
     n = rb_scan_args(argc, argv, "03", &args[1], &args[2], &args[3]);
-    return rb_funcallv(rb_cIO, rb_intern("binwrite"), 1+n, args);
+    return rb_funcallv_kw(rb_cFile, id_binwrite, 1+n, args, RB_PASS_CALLED_KEYWORDS);
 }
 
 /*
@@ -394,7 +466,7 @@ path_binwrite(int argc, VALUE *argv, VALUE self)
  *
  * Returns all the lines from the file.
  *
- * See IO.readlines.
+ * See File.readlines.
  *
  */
 static VALUE
@@ -405,7 +477,7 @@ path_readlines(int argc, VALUE *argv, VALUE self)
 
     args[0] = get_strpath(self);
     n = rb_scan_args(argc, argv, "03", &args[1], &args[2], &args[3]);
-    return rb_funcallv(rb_cIO, rb_intern("readlines"), 1+n, args);
+    return rb_funcallv_kw(rb_cFile, id_readlines, 1+n, args, RB_PASS_CALLED_KEYWORDS);
 }
 
 /*
@@ -423,7 +495,7 @@ path_sysopen(int argc, VALUE *argv, VALUE self)
 
     args[0] = get_strpath(self);
     n = rb_scan_args(argc, argv, "02", &args[1], &args[2]);
-    return rb_funcallv(rb_cIO, rb_intern("sysopen"), 1+n, args);
+    return rb_funcallv(rb_cIO, id_sysopen, 1+n, args);
 }
 
 /*
@@ -437,10 +509,10 @@ path_sysopen(int argc, VALUE *argv, VALUE self)
 static VALUE
 path_atime(VALUE self)
 {
-    return rb_funcall(rb_cFile, rb_intern("atime"), 1, get_strpath(self));
+    return rb_funcall(rb_cFile, id_atime, 1, get_strpath(self));
 }
 
-#if defined(HAVE_STRUCT_STAT_ST_BIRTHTIMESPEC) || defined(_WIN32)
+#if defined(HAVE_RB_FILE_S_BIRTHTIME)
 /*
  * call-seq:
  *   pathname.birthtime	-> time
@@ -453,9 +525,10 @@ path_atime(VALUE self)
 static VALUE
 path_birthtime(VALUE self)
 {
-    return rb_funcall(rb_cFile, rb_intern("birthtime"), 1, get_strpath(self));
+    return rb_funcall(rb_cFile, id_birthtime, 1, get_strpath(self));
 }
 #else
+/* check at compilation time for `respond_to?` */
 # define path_birthtime rb_f_notimplement
 #endif
 
@@ -470,7 +543,7 @@ path_birthtime(VALUE self)
 static VALUE
 path_ctime(VALUE self)
 {
-    return rb_funcall(rb_cFile, rb_intern("ctime"), 1, get_strpath(self));
+    return rb_funcall(rb_cFile, id_ctime, 1, get_strpath(self));
 }
 
 /*
@@ -484,7 +557,7 @@ path_ctime(VALUE self)
 static VALUE
 path_mtime(VALUE self)
 {
-    return rb_funcall(rb_cFile, rb_intern("mtime"), 1, get_strpath(self));
+    return rb_funcall(rb_cFile, id_mtime, 1, get_strpath(self));
 }
 
 /*
@@ -498,7 +571,7 @@ path_mtime(VALUE self)
 static VALUE
 path_chmod(VALUE self, VALUE mode)
 {
-    return rb_funcall(rb_cFile, rb_intern("chmod"), 2, mode, get_strpath(self));
+    return rb_funcall(rb_cFile, id_chmod, 2, mode, get_strpath(self));
 }
 
 /*
@@ -512,7 +585,7 @@ path_chmod(VALUE self, VALUE mode)
 static VALUE
 path_lchmod(VALUE self, VALUE mode)
 {
-    return rb_funcall(rb_cFile, rb_intern("lchmod"), 2, mode, get_strpath(self));
+    return rb_funcall(rb_cFile, id_lchmod, 2, mode, get_strpath(self));
 }
 
 /*
@@ -526,7 +599,7 @@ path_lchmod(VALUE self, VALUE mode)
 static VALUE
 path_chown(VALUE self, VALUE owner, VALUE group)
 {
-    return rb_funcall(rb_cFile, rb_intern("chown"), 3, owner, group, get_strpath(self));
+    return rb_funcall(rb_cFile, id_chown, 3, owner, group, get_strpath(self));
 }
 
 /*
@@ -540,7 +613,7 @@ path_chown(VALUE self, VALUE owner, VALUE group)
 static VALUE
 path_lchown(VALUE self, VALUE owner, VALUE group)
 {
-    return rb_funcall(rb_cFile, rb_intern("lchown"), 3, owner, group, get_strpath(self));
+    return rb_funcall(rb_cFile, id_lchown, 3, owner, group, get_strpath(self));
 }
 
 /*
@@ -558,9 +631,9 @@ path_fnmatch(int argc, VALUE *argv, VALUE self)
     VALUE str = get_strpath(self);
     VALUE pattern, flags;
     if (rb_scan_args(argc, argv, "11", &pattern, &flags) == 1)
-        return rb_funcall(rb_cFile, rb_intern("fnmatch"), 2, pattern, str);
+        return rb_funcall(rb_cFile, id_fnmatch, 2, pattern, str);
     else
-        return rb_funcall(rb_cFile, rb_intern("fnmatch"), 3, pattern, str, flags);
+        return rb_funcall(rb_cFile, id_fnmatch, 3, pattern, str, flags);
 }
 
 /*
@@ -574,7 +647,7 @@ path_fnmatch(int argc, VALUE *argv, VALUE self)
 static VALUE
 path_ftype(VALUE self)
 {
-    return rb_funcall(rb_cFile, rb_intern("ftype"), 1, get_strpath(self));
+    return rb_funcall(rb_cFile, id_ftype, 1, get_strpath(self));
 }
 
 /*
@@ -588,7 +661,7 @@ path_ftype(VALUE self)
 static VALUE
 path_make_link(VALUE self, VALUE old)
 {
-    return rb_funcall(rb_cFile, rb_intern("link"), 2, old, get_strpath(self));
+    return rb_funcall(rb_cFile, id_link, 2, old, get_strpath(self));
 }
 
 /*
@@ -605,10 +678,10 @@ path_open(int argc, VALUE *argv, VALUE self)
     args[0] = get_strpath(self);
     n = rb_scan_args(argc, argv, "03", &args[1], &args[2], &args[3]);
     if (rb_block_given_p()) {
-        return rb_block_call(rb_cFile, rb_intern("open"), 1+n, args, 0, 0);
+        return rb_block_call_kw(rb_cFile, id_open, 1+n, args, 0, 0, RB_PASS_CALLED_KEYWORDS);
     }
     else {
-        return rb_funcallv(rb_cFile, rb_intern("open"), 1+n, args);
+        return rb_funcallv_kw(rb_cFile, id_open, 1+n, args, RB_PASS_CALLED_KEYWORDS);
     }
 }
 
@@ -621,7 +694,7 @@ static VALUE
 path_readlink(VALUE self)
 {
     VALUE str;
-    str = rb_funcall(rb_cFile, rb_intern("readlink"), 1, get_strpath(self));
+    str = rb_funcall(rb_cFile, id_readlink, 1, get_strpath(self));
     return rb_class_new_instance(1, &str, rb_obj_class(self));
 }
 
@@ -633,7 +706,7 @@ path_readlink(VALUE self)
 static VALUE
 path_rename(VALUE self, VALUE to)
 {
-    return rb_funcall(rb_cFile, rb_intern("rename"), 2, get_strpath(self), to);
+    return rb_funcall(rb_cFile, id_rename, 2, get_strpath(self), to);
 }
 
 /*
@@ -644,7 +717,7 @@ path_rename(VALUE self, VALUE to)
 static VALUE
 path_stat(VALUE self)
 {
-    return rb_funcall(rb_cFile, rb_intern("stat"), 1, get_strpath(self));
+    return rb_funcall(rb_cFile, id_stat, 1, get_strpath(self));
 }
 
 /*
@@ -653,7 +726,7 @@ path_stat(VALUE self)
 static VALUE
 path_lstat(VALUE self)
 {
-    return rb_funcall(rb_cFile, rb_intern("lstat"), 1, get_strpath(self));
+    return rb_funcall(rb_cFile, id_lstat, 1, get_strpath(self));
 }
 
 /*
@@ -667,7 +740,7 @@ path_lstat(VALUE self)
 static VALUE
 path_make_symlink(VALUE self, VALUE old)
 {
-    return rb_funcall(rb_cFile, rb_intern("symlink"), 2, old, get_strpath(self));
+    return rb_funcall(rb_cFile, id_symlink, 2, old, get_strpath(self));
 }
 
 /*
@@ -678,7 +751,7 @@ path_make_symlink(VALUE self, VALUE old)
 static VALUE
 path_truncate(VALUE self, VALUE length)
 {
-    return rb_funcall(rb_cFile, rb_intern("truncate"), 2, get_strpath(self), length);
+    return rb_funcall(rb_cFile, id_truncate, 2, get_strpath(self), length);
 }
 
 /*
@@ -689,7 +762,7 @@ path_truncate(VALUE self, VALUE length)
 static VALUE
 path_utime(VALUE self, VALUE atime, VALUE mtime)
 {
-    return rb_funcall(rb_cFile, rb_intern("utime"), 3, atime, mtime, get_strpath(self));
+    return rb_funcall(rb_cFile, id_utime, 3, atime, mtime, get_strpath(self));
 }
 
 /*
@@ -703,9 +776,9 @@ path_basename(int argc, VALUE *argv, VALUE self)
     VALUE str = get_strpath(self);
     VALUE fext;
     if (rb_scan_args(argc, argv, "01", &fext) == 0)
-        str = rb_funcall(rb_cFile, rb_intern("basename"), 1, str);
+        str = rb_funcall(rb_cFile, id_basename, 1, str);
     else
-        str = rb_funcall(rb_cFile, rb_intern("basename"), 2, str, fext);
+        str = rb_funcall(rb_cFile, id_basename, 2, str, fext);
     return rb_class_new_instance(1, &str, rb_obj_class(self));
 }
 
@@ -718,7 +791,7 @@ static VALUE
 path_dirname(VALUE self)
 {
     VALUE str = get_strpath(self);
-    str = rb_funcall(rb_cFile, rb_intern("dirname"), 1, str);
+    str = rb_funcall(rb_cFile, id_dirname, 1, str);
     return rb_class_new_instance(1, &str, rb_obj_class(self));
 }
 
@@ -731,7 +804,7 @@ static VALUE
 path_extname(VALUE self)
 {
     VALUE str = get_strpath(self);
-    return rb_funcall(rb_cFile, rb_intern("extname"), 1, str);
+    return rb_funcall(rb_cFile, id_extname, 1, str);
 }
 
 /*
@@ -745,9 +818,9 @@ path_expand_path(int argc, VALUE *argv, VALUE self)
     VALUE str = get_strpath(self);
     VALUE dname;
     if (rb_scan_args(argc, argv, "01", &dname) == 0)
-        str = rb_funcall(rb_cFile, rb_intern("expand_path"), 1, str);
+        str = rb_funcall(rb_cFile, id_expand_path, 1, str);
     else
-        str = rb_funcall(rb_cFile, rb_intern("expand_path"), 2, str, dname);
+        str = rb_funcall(rb_cFile, id_expand_path, 2, str, dname);
     return rb_class_new_instance(1, &str, rb_obj_class(self));
 }
 
@@ -761,7 +834,7 @@ path_split(VALUE self)
 {
     VALUE str = get_strpath(self);
     VALUE ary, dirname, basename;
-    ary = rb_funcall(rb_cFile, rb_intern("split"), 1, str);
+    ary = rb_funcall(rb_cFile, id_split, 1, str);
     ary = rb_check_array_type(ary);
     dirname = rb_ary_entry(ary, 0);
     basename = rb_ary_entry(ary, 1);
@@ -776,7 +849,7 @@ path_split(VALUE self)
 static VALUE
 path_blockdev_p(VALUE self)
 {
-    return rb_funcall(rb_mFileTest, rb_intern("blockdev?"), 1, get_strpath(self));
+    return rb_funcall(rb_mFileTest, id_blockdev_p, 1, get_strpath(self));
 }
 
 /*
@@ -785,7 +858,7 @@ path_blockdev_p(VALUE self)
 static VALUE
 path_chardev_p(VALUE self)
 {
-    return rb_funcall(rb_mFileTest, rb_intern("chardev?"), 1, get_strpath(self));
+    return rb_funcall(rb_mFileTest, id_chardev_p, 1, get_strpath(self));
 }
 
 /*
@@ -794,7 +867,7 @@ path_chardev_p(VALUE self)
 static VALUE
 path_executable_p(VALUE self)
 {
-    return rb_funcall(rb_mFileTest, rb_intern("executable?"), 1, get_strpath(self));
+    return rb_funcall(rb_mFileTest, id_executable_p, 1, get_strpath(self));
 }
 
 /*
@@ -803,7 +876,7 @@ path_executable_p(VALUE self)
 static VALUE
 path_executable_real_p(VALUE self)
 {
-    return rb_funcall(rb_mFileTest, rb_intern("executable_real?"), 1, get_strpath(self));
+    return rb_funcall(rb_mFileTest, id_executable_real_p, 1, get_strpath(self));
 }
 
 /*
@@ -812,7 +885,7 @@ path_executable_real_p(VALUE self)
 static VALUE
 path_exist_p(VALUE self)
 {
-    return rb_funcall(rb_mFileTest, rb_intern("exist?"), 1, get_strpath(self));
+    return rb_funcall(rb_mFileTest, id_exist_p, 1, get_strpath(self));
 }
 
 /*
@@ -821,7 +894,7 @@ path_exist_p(VALUE self)
 static VALUE
 path_grpowned_p(VALUE self)
 {
-    return rb_funcall(rb_mFileTest, rb_intern("grpowned?"), 1, get_strpath(self));
+    return rb_funcall(rb_mFileTest, id_grpowned_p, 1, get_strpath(self));
 }
 
 /*
@@ -830,7 +903,7 @@ path_grpowned_p(VALUE self)
 static VALUE
 path_directory_p(VALUE self)
 {
-    return rb_funcall(rb_mFileTest, rb_intern("directory?"), 1, get_strpath(self));
+    return rb_funcall(rb_mFileTest, id_directory_p, 1, get_strpath(self));
 }
 
 /*
@@ -839,7 +912,7 @@ path_directory_p(VALUE self)
 static VALUE
 path_file_p(VALUE self)
 {
-    return rb_funcall(rb_mFileTest, rb_intern("file?"), 1, get_strpath(self));
+    return rb_funcall(rb_mFileTest, id_file_p, 1, get_strpath(self));
 }
 
 /*
@@ -848,7 +921,7 @@ path_file_p(VALUE self)
 static VALUE
 path_pipe_p(VALUE self)
 {
-    return rb_funcall(rb_mFileTest, rb_intern("pipe?"), 1, get_strpath(self));
+    return rb_funcall(rb_mFileTest, id_pipe_p, 1, get_strpath(self));
 }
 
 /*
@@ -857,7 +930,7 @@ path_pipe_p(VALUE self)
 static VALUE
 path_socket_p(VALUE self)
 {
-    return rb_funcall(rb_mFileTest, rb_intern("socket?"), 1, get_strpath(self));
+    return rb_funcall(rb_mFileTest, id_socket_p, 1, get_strpath(self));
 }
 
 /*
@@ -866,7 +939,7 @@ path_socket_p(VALUE self)
 static VALUE
 path_owned_p(VALUE self)
 {
-    return rb_funcall(rb_mFileTest, rb_intern("owned?"), 1, get_strpath(self));
+    return rb_funcall(rb_mFileTest, id_owned_p, 1, get_strpath(self));
 }
 
 /*
@@ -875,7 +948,7 @@ path_owned_p(VALUE self)
 static VALUE
 path_readable_p(VALUE self)
 {
-    return rb_funcall(rb_mFileTest, rb_intern("readable?"), 1, get_strpath(self));
+    return rb_funcall(rb_mFileTest, id_readable_p, 1, get_strpath(self));
 }
 
 /*
@@ -884,7 +957,7 @@ path_readable_p(VALUE self)
 static VALUE
 path_world_readable_p(VALUE self)
 {
-    return rb_funcall(rb_mFileTest, rb_intern("world_readable?"), 1, get_strpath(self));
+    return rb_funcall(rb_mFileTest, id_world_readable_p, 1, get_strpath(self));
 }
 
 /*
@@ -893,7 +966,7 @@ path_world_readable_p(VALUE self)
 static VALUE
 path_readable_real_p(VALUE self)
 {
-    return rb_funcall(rb_mFileTest, rb_intern("readable_real?"), 1, get_strpath(self));
+    return rb_funcall(rb_mFileTest, id_readable_real_p, 1, get_strpath(self));
 }
 
 /*
@@ -902,7 +975,7 @@ path_readable_real_p(VALUE self)
 static VALUE
 path_setuid_p(VALUE self)
 {
-    return rb_funcall(rb_mFileTest, rb_intern("setuid?"), 1, get_strpath(self));
+    return rb_funcall(rb_mFileTest, id_setuid_p, 1, get_strpath(self));
 }
 
 /*
@@ -911,7 +984,7 @@ path_setuid_p(VALUE self)
 static VALUE
 path_setgid_p(VALUE self)
 {
-    return rb_funcall(rb_mFileTest, rb_intern("setgid?"), 1, get_strpath(self));
+    return rb_funcall(rb_mFileTest, id_setgid_p, 1, get_strpath(self));
 }
 
 /*
@@ -920,7 +993,7 @@ path_setgid_p(VALUE self)
 static VALUE
 path_size(VALUE self)
 {
-    return rb_funcall(rb_mFileTest, rb_intern("size"), 1, get_strpath(self));
+    return rb_funcall(rb_mFileTest, id_size, 1, get_strpath(self));
 }
 
 /*
@@ -929,7 +1002,7 @@ path_size(VALUE self)
 static VALUE
 path_size_p(VALUE self)
 {
-    return rb_funcall(rb_mFileTest, rb_intern("size?"), 1, get_strpath(self));
+    return rb_funcall(rb_mFileTest, id_size_p, 1, get_strpath(self));
 }
 
 /*
@@ -938,7 +1011,7 @@ path_size_p(VALUE self)
 static VALUE
 path_sticky_p(VALUE self)
 {
-    return rb_funcall(rb_mFileTest, rb_intern("sticky?"), 1, get_strpath(self));
+    return rb_funcall(rb_mFileTest, id_sticky_p, 1, get_strpath(self));
 }
 
 /*
@@ -947,7 +1020,7 @@ path_sticky_p(VALUE self)
 static VALUE
 path_symlink_p(VALUE self)
 {
-    return rb_funcall(rb_mFileTest, rb_intern("symlink?"), 1, get_strpath(self));
+    return rb_funcall(rb_mFileTest, id_symlink_p, 1, get_strpath(self));
 }
 
 /*
@@ -956,7 +1029,7 @@ path_symlink_p(VALUE self)
 static VALUE
 path_writable_p(VALUE self)
 {
-    return rb_funcall(rb_mFileTest, rb_intern("writable?"), 1, get_strpath(self));
+    return rb_funcall(rb_mFileTest, id_writable_p, 1, get_strpath(self));
 }
 
 /*
@@ -965,7 +1038,7 @@ path_writable_p(VALUE self)
 static VALUE
 path_world_writable_p(VALUE self)
 {
-    return rb_funcall(rb_mFileTest, rb_intern("world_writable?"), 1, get_strpath(self));
+    return rb_funcall(rb_mFileTest, id_world_writable_p, 1, get_strpath(self));
 }
 
 /*
@@ -974,7 +1047,7 @@ path_world_writable_p(VALUE self)
 static VALUE
 path_writable_real_p(VALUE self)
 {
-    return rb_funcall(rb_mFileTest, rb_intern("writable_real?"), 1, get_strpath(self));
+    return rb_funcall(rb_mFileTest, id_writable_real_p, 1, get_strpath(self));
 }
 
 /*
@@ -983,11 +1056,27 @@ path_writable_real_p(VALUE self)
 static VALUE
 path_zero_p(VALUE self)
 {
-    return rb_funcall(rb_mFileTest, rb_intern("zero?"), 1, get_strpath(self));
+    return rb_funcall(rb_mFileTest, id_zero_p, 1, get_strpath(self));
+}
+
+/*
+ * Tests the file is empty.
+ *
+ * See Dir#empty? and FileTest.empty?.
+ */
+static VALUE
+path_empty_p(VALUE self)
+{
+
+    VALUE path = get_strpath(self);
+    if (RTEST(rb_funcall(rb_mFileTest, id_directory_p, 1, path)))
+        return rb_funcall(rb_cDir, id_empty_p, 1, path);
+    else
+        return rb_funcall(rb_mFileTest, id_empty_p, 1, path);
 }
 
 static VALUE
-glob_i(RB_BLOCK_CALL_FUNC_ARGLIST(elt, klass))
+s_glob_i(RB_BLOCK_CALL_FUNC_ARGLIST(elt, klass))
 {
     return rb_yield(rb_class_new_instance(1, &elt, klass));
 }
@@ -995,25 +1084,25 @@ glob_i(RB_BLOCK_CALL_FUNC_ARGLIST(elt, klass))
 /*
  * Returns or yields Pathname objects.
  *
- *  Pathname.glob("config/" "*.rb")
- *	#=> [#<Pathname:config/environment.rb>, #<Pathname:config/routes.rb>, ..]
+ *  Pathname.glob("lib/i*.rb")
+ *	#=> [#<Pathname:lib/ipaddr.rb>, #<Pathname:lib/irb.rb>]
  *
  * See Dir.glob.
  */
 static VALUE
 path_s_glob(int argc, VALUE *argv, VALUE klass)
 {
-    VALUE args[2];
+    VALUE args[3];
     int n;
 
-    n = rb_scan_args(argc, argv, "11", &args[0], &args[1]);
+    n = rb_scan_args(argc, argv, "12", &args[0], &args[1], &args[2]);
     if (rb_block_given_p()) {
-        return rb_block_call(rb_cDir, rb_intern("glob"), n, args, glob_i, klass);
+        return rb_block_call_kw(rb_cDir, id_glob, n, args, s_glob_i, klass, RB_PASS_CALLED_KEYWORDS);
     }
     else {
         VALUE ary;
         long i;
-        ary = rb_funcallv(rb_cDir, rb_intern("glob"), n, args);
+        ary = rb_funcallv_kw(rb_cDir, id_glob, n, args, RB_PASS_CALLED_KEYWORDS);
         ary = rb_convert_type(ary, T_ARRAY, "Array", "to_ary");
         for (i = 0; i < RARRAY_LEN(ary); i++) {
             VALUE elt = RARRAY_AREF(ary, i);
@@ -1024,6 +1113,54 @@ path_s_glob(int argc, VALUE *argv, VALUE klass)
     }
 }
 
+static VALUE
+glob_i(RB_BLOCK_CALL_FUNC_ARGLIST(elt, self))
+{
+    elt = rb_funcall(self, '+', 1, elt);
+    return rb_yield(elt);
+}
+
+/*
+ * Returns or yields Pathname objects.
+ *
+ *  Pathname("ruby-2.4.2").glob("R*.md")
+ *  #=> [#<Pathname:ruby-2.4.2/README.md>, #<Pathname:ruby-2.4.2/README.ja.md>]
+ *
+ * See Dir.glob.
+ * This method uses the +base+ keyword argument of Dir.glob.
+ */
+static VALUE
+path_glob(int argc, VALUE *argv, VALUE self)
+{
+    VALUE args[3];
+    int n;
+
+    n = rb_scan_args(argc, argv, "11", &args[0], &args[1]);
+    if (n == 1)
+      args[1] = INT2FIX(0);
+
+    args[2] = rb_hash_new();
+    rb_hash_aset(args[2], ID2SYM(id_base), get_strpath(self));
+
+    n = 3;
+
+    if (rb_block_given_p()) {
+        return rb_block_call_kw(rb_cDir, id_glob, n, args, glob_i, self, RB_PASS_KEYWORDS);
+    }
+    else {
+        VALUE ary;
+        long i;
+        ary = rb_funcallv_kw(rb_cDir, id_glob, n, args, RB_PASS_KEYWORDS);
+        ary = rb_convert_type(ary, T_ARRAY, "Array", "to_ary");
+        for (i = 0; i < RARRAY_LEN(ary); i++) {
+            VALUE elt = RARRAY_AREF(ary, i);
+            elt = rb_funcall(self, '+', 1, elt);
+            rb_ary_store(ary, i, elt);
+        }
+        return ary;
+    }
+}
+
 /*
  * Returns the current working directory as a Pathname.
  *
@@ -1036,7 +1173,7 @@ static VALUE
 path_s_getwd(VALUE klass)
 {
     VALUE str;
-    str = rb_funcall(rb_cDir, rb_intern("getwd"), 0);
+    str = rb_funcall(rb_cDir, id_getwd, 0);
     return rb_class_new_instance(1, &str, klass);
 }
 
@@ -1073,7 +1210,7 @@ path_entries(VALUE self)
     long i;
     klass = rb_obj_class(self);
     str = get_strpath(self);
-    ary = rb_funcall(rb_cDir, rb_intern("entries"), 1, str);
+    ary = rb_funcall(rb_cDir, id_entries, 1, str);
     ary = rb_convert_type(ary, T_ARRAY, "Array", "to_ary");
     for (i = 0; i < RARRAY_LEN(ary); i++) {
 	VALUE elt = RARRAY_AREF(ary, i);
@@ -1094,9 +1231,9 @@ path_mkdir(int argc, VALUE *argv, VALUE self)
     VALUE str = get_strpath(self);
     VALUE vmode;
     if (rb_scan_args(argc, argv, "01", &vmode) == 0)
-        return rb_funcall(rb_cDir, rb_intern("mkdir"), 1, str);
+        return rb_funcall(rb_cDir, id_mkdir, 1, str);
     else
-        return rb_funcall(rb_cDir, rb_intern("mkdir"), 2, str, vmode);
+        return rb_funcall(rb_cDir, id_mkdir, 2, str, vmode);
 }
 
 /*
@@ -1107,7 +1244,7 @@ path_mkdir(int argc, VALUE *argv, VALUE self)
 static VALUE
 path_rmdir(VALUE self)
 {
-    return rb_funcall(rb_cDir, rb_intern("rmdir"), 1, get_strpath(self));
+    return rb_funcall(rb_cDir, id_rmdir, 1, get_strpath(self));
 }
 
 /*
@@ -1121,7 +1258,7 @@ path_opendir(VALUE self)
     VALUE args[1];
 
     args[0] = get_strpath(self);
-    return rb_block_call(rb_cDir, rb_intern("open"), 1, args, 0, 0);
+    return rb_block_call(rb_cDir, id_open, 1, args, 0, 0);
 }
 
 static VALUE
@@ -1140,19 +1277,19 @@ path_each_entry(VALUE self)
     VALUE args[1];
 
     args[0] = get_strpath(self);
-    return rb_block_call(rb_cDir, rb_intern("foreach"), 1, args, each_entry_i, rb_obj_class(self));
+    return rb_block_call(rb_cDir, id_foreach, 1, args, each_entry_i, rb_obj_class(self));
 }
 
 static VALUE
 unlink_body(VALUE str)
 {
-    return rb_funcall(rb_cDir, rb_intern("unlink"), 1, str);
+    return rb_funcall(rb_cDir, id_unlink, 1, str);
 }
 
 static VALUE
 unlink_rescue(VALUE str, VALUE errinfo)
 {
-    return rb_funcall(rb_cFile, rb_intern("unlink"), 1, str);
+    return rb_funcall(rb_cFile, id_unlink, 1, str);
 }
 
 /*
@@ -1162,7 +1299,7 @@ unlink_rescue(VALUE str, VALUE errinfo)
 static VALUE
 path_unlink(VALUE self)
 {
-    VALUE eENOTDIR = rb_const_get_at(rb_mErrno, rb_intern("ENOTDIR"));
+    VALUE eENOTDIR = rb_const_get_at(rb_mErrno, id_ENOTDIR);
     VALUE str = get_strpath(self);
     return rb_rescue2(unlink_body, str, unlink_rescue, str, eENOTDIR, (VALUE)0);
 }
@@ -1186,6 +1323,8 @@ path_unlink(VALUE self)
 static VALUE
 path_f_pathname(VALUE self, VALUE str)
 {
+    if (CLASS_OF(str) == rb_cPathname)
+        return str;
     return rb_class_new_instance(1, &str, rb_cPathname);
 }
 
@@ -1372,8 +1511,7 @@ path_f_pathname(VALUE self, VALUE str)
 void
 Init_pathname(void)
 {
-    id_at_path = rb_intern("@path");
-    id_to_path = rb_intern("to_path");
+    InitVM(pathname);
 
     rb_cPathname = rb_define_class("Pathname", rb_cObject);
     rb_define_method(rb_cPathname, "initialize", path_initialize, 1);
@@ -1448,9 +1586,11 @@ Init_pathname(void)
     rb_define_method(rb_cPathname, "world_writable?", path_world_writable_p, 0);
     rb_define_method(rb_cPathname, "writable_real?", path_writable_real_p, 0);
     rb_define_method(rb_cPathname, "zero?", path_zero_p, 0);
+    rb_define_method(rb_cPathname, "empty?", path_empty_p, 0);
     rb_define_singleton_method(rb_cPathname, "glob", path_s_glob, -1);
     rb_define_singleton_method(rb_cPathname, "getwd", path_s_getwd, 0);
     rb_define_singleton_method(rb_cPathname, "pwd", path_s_getwd, 0);
+    rb_define_method(rb_cPathname, "glob", path_glob, -1);
     rb_define_method(rb_cPathname, "entries", path_entries, 0);
     rb_define_method(rb_cPathname, "mkdir", path_mkdir, -1);
     rb_define_method(rb_cPathname, "rmdir", path_rmdir, 0);
@@ -1461,3 +1601,78 @@ Init_pathname(void)
     rb_undef_method(rb_cPathname, "=~");
     rb_define_global_function("Pathname", path_f_pathname, 1);
 }
+
+void
+InitVM_pathname(void)
+{
+#undef rb_intern
+    id_at_path = rb_intern("@path");
+    id_to_path = rb_intern("to_path");
+    id_ENOTDIR = rb_intern("ENOTDIR");
+    id_atime = rb_intern("atime");
+    id_basename = rb_intern("basename");
+    id_base = rb_intern("base");
+    id_binread = rb_intern("binread");
+    id_binwrite = rb_intern("binwrite");
+    id_birthtime = rb_intern("birthtime");
+    id_blockdev_p = rb_intern("blockdev?");
+    id_chardev_p = rb_intern("chardev?");
+    id_chmod = rb_intern("chmod");
+    id_chown = rb_intern("chown");
+    id_ctime = rb_intern("ctime");
+    id_directory_p = rb_intern("directory?");
+    id_dirname = rb_intern("dirname");
+    id_empty_p = rb_intern("empty?");
+    id_entries = rb_intern("entries");
+    id_executable_p = rb_intern("executable?");
+    id_executable_real_p = rb_intern("executable_real?");
+    id_exist_p = rb_intern("exist?");
+    id_expand_path = rb_intern("expand_path");
+    id_extname = rb_intern("extname");
+    id_file_p = rb_intern("file?");
+    id_fnmatch = rb_intern("fnmatch");
+    id_foreach = rb_intern("foreach");
+    id_ftype = rb_intern("ftype");
+    id_getwd = rb_intern("getwd");
+    id_glob = rb_intern("glob");
+    id_grpowned_p = rb_intern("grpowned?");
+    id_lchmod = rb_intern("lchmod");
+    id_lchown = rb_intern("lchown");
+    id_link = rb_intern("link");
+    id_lstat = rb_intern("lstat");
+    id_mkdir = rb_intern("mkdir");
+    id_mtime = rb_intern("mtime");
+    id_open = rb_intern("open");
+    id_owned_p = rb_intern("owned?");
+    id_pipe_p = rb_intern("pipe?");
+    id_read = rb_intern("read");
+    id_readable_p = rb_intern("readable?");
+    id_readable_real_p = rb_intern("readable_real?");
+    id_readlines = rb_intern("readlines");
+    id_readlink = rb_intern("readlink");
+    id_realdirpath = rb_intern("realdirpath");
+    id_realpath = rb_intern("realpath");
+    id_rename = rb_intern("rename");
+    id_rmdir = rb_intern("rmdir");
+    id_setgid_p = rb_intern("setgid?");
+    id_setuid_p = rb_intern("setuid?");
+    id_size = rb_intern("size");
+    id_size_p = rb_intern("size?");
+    id_socket_p = rb_intern("socket?");
+    id_split = rb_intern("split");
+    id_stat = rb_intern("stat");
+    id_sticky_p = rb_intern("sticky?");
+    id_sub = rb_intern("sub");
+    id_symlink = rb_intern("symlink");
+    id_symlink_p = rb_intern("symlink?");
+    id_sysopen = rb_intern("sysopen");
+    id_truncate = rb_intern("truncate");
+    id_unlink = rb_intern("unlink");
+    id_utime = rb_intern("utime");
+    id_world_readable_p = rb_intern("world_readable?");
+    id_world_writable_p = rb_intern("world_writable?");
+    id_writable_p = rb_intern("writable?");
+    id_writable_real_p = rb_intern("writable_real?");
+    id_write = rb_intern("write");
+    id_zero_p = rb_intern("zero?");
+}
