@@ -16,12 +16,12 @@
 # NOTE: You can find Japanese version of this document at:
 # http://docs.ruby-lang.org/ja/latest/library/net=2fpop.html
 #
-#   $Id: pop.rb 62099 2018-01-30 00:22:22Z normal $
+#   $Id$
 #
 # See Net::POP3 for documentation.
 #
 
-require 'net/protocol'
+require_relative 'protocol'
 require 'digest/md5'
 require 'timeout'
 
@@ -196,7 +196,7 @@ module Net
   class POP3 < Protocol
 
     # svn revision of this library
-    Revision = %q$Revision: 62099 $.split[1]
+    Revision = %q$Revision$.split[1]
 
     #
     # Class Parameters
@@ -549,6 +549,7 @@ module Net
         context = OpenSSL::SSL::SSLContext.new
         context.set_params(@ssl_params)
         s = OpenSSL::SSL::SSLSocket.new(s, context)
+        s.hostname = @address
         s.sync_close = true
         ssl_socket_connect(s, @open_timeout)
         if context.verify_mode != OpenSSL::SSL::VERIFY_NONE
