@@ -230,7 +230,7 @@ describe "CApiObject" do
       obj = mock("rb_check_convert_type")
       obj.should_receive(:to_array).and_return("string")
 
-      lambda do
+      -> do
         @o.rb_check_convert_type(obj, "Array", "to_array")
       end.should raise_error(TypeError)
     end
@@ -255,7 +255,7 @@ describe "CApiObject" do
       obj = mock("rb_convert_type")
       obj.should_receive(:to_array).and_return(nil)
 
-      lambda do
+      -> do
         @o.rb_convert_type(obj, "Array", "to_array")
       end.should raise_error(TypeError)
     end
@@ -264,7 +264,7 @@ describe "CApiObject" do
       obj = mock("rb_convert_type")
       obj.should_receive(:to_array).and_return("string")
 
-      lambda do
+      -> do
         @o.rb_convert_type(obj, "Array", "to_array")
       end.should raise_error(TypeError)
     end
@@ -308,13 +308,13 @@ describe "CApiObject" do
     it "sends #to_ary to the argument and raises TypeError if it's not a kind of Array" do
       obj = mock("to_ary")
       obj.should_receive(:to_ary).and_return(Object.new)
-      lambda { @o.rb_check_array_type obj }.should raise_error(TypeError)
+      -> { @o.rb_check_array_type obj }.should raise_error(TypeError)
     end
 
     it "does not rescue exceptions raised by #to_ary" do
       obj = mock("to_ary")
       obj.should_receive(:to_ary).and_raise(frozen_error_class)
-      lambda { @o.rb_check_array_type obj }.should raise_error(frozen_error_class)
+      -> { @o.rb_check_array_type obj }.should raise_error(frozen_error_class)
     end
   end
 
@@ -356,13 +356,13 @@ describe "CApiObject" do
     it "sends #to_str to the argument and raises TypeError if it's not a kind of String" do
       obj = mock("to_str")
       obj.should_receive(:to_str).and_return(Object.new)
-      lambda { @o.rb_check_string_type obj }.should raise_error(TypeError)
+      -> { @o.rb_check_string_type obj }.should raise_error(TypeError)
     end
 
     it "does not rescue exceptions raised by #to_str" do
       obj = mock("to_str")
       obj.should_receive(:to_str).and_raise(RuntimeError)
-      lambda { @o.rb_check_string_type obj }.should raise_error(RuntimeError)
+      -> { @o.rb_check_string_type obj }.should raise_error(RuntimeError)
     end
   end
 
@@ -639,7 +639,7 @@ describe "CApiObject" do
   describe "rb_obj_instance_eval" do
     it "evaluates the block in the object context, that includes private methods" do
       obj = ObjectTest
-      lambda do
+      -> do
         @o.rb_obj_instance_eval(obj) { include Kernel }
       end.should_not raise_error(NoMethodError)
     end
@@ -667,18 +667,18 @@ describe "CApiObject" do
     end
 
     it "raises a #{frozen_error_class} if the object passed is frozen" do
-      lambda { @o.rb_obj_taint("".freeze) }.should raise_error(frozen_error_class)
+      -> { @o.rb_obj_taint("".freeze) }.should raise_error(frozen_error_class)
     end
   end
 
   describe "rb_check_frozen" do
     it "raises a #{frozen_error_class} if the obj is frozen" do
-      lambda { @o.rb_check_frozen("".freeze) }.should raise_error(frozen_error_class)
+      -> { @o.rb_check_frozen("".freeze) }.should raise_error(frozen_error_class)
     end
 
     it "does nothing when object isn't frozen" do
       obj = ""
-      lambda { @o.rb_check_frozen(obj) }.should_not raise_error(TypeError)
+      -> { @o.rb_check_frozen(obj) }.should_not raise_error(TypeError)
     end
   end
 
@@ -718,23 +718,23 @@ describe "CApiObject" do
     it "raises a TypeError if #to_int does not return an Integer" do
       x = mock("to_int")
       x.should_receive(:to_int).and_return("5")
-      lambda { @o.rb_to_int(x) }.should raise_error(TypeError)
+      -> { @o.rb_to_int(x) }.should raise_error(TypeError)
     end
 
     it "raises a TypeError if called with nil" do
-      lambda { @o.rb_to_int(nil) }.should raise_error(TypeError)
+      -> { @o.rb_to_int(nil) }.should raise_error(TypeError)
     end
 
     it "raises a TypeError if called with true" do
-      lambda { @o.rb_to_int(true) }.should raise_error(TypeError)
+      -> { @o.rb_to_int(true) }.should raise_error(TypeError)
     end
 
     it "raises a TypeError if called with false" do
-      lambda { @o.rb_to_int(false) }.should raise_error(TypeError)
+      -> { @o.rb_to_int(false) }.should raise_error(TypeError)
     end
 
     it "raises a TypeError if called with a String" do
-      lambda { @o.rb_to_int("1") }.should raise_error(TypeError)
+      -> { @o.rb_to_int("1") }.should raise_error(TypeError)
     end
   end
 
@@ -770,7 +770,7 @@ describe "CApiObject" do
     end
 
     it "raises a TypeError if arg is no class or module" do
-      lambda{
+      ->{
         @o.rb_class_inherited_p(1, 2)
       }.should raise_error(TypeError)
     end
@@ -813,6 +813,15 @@ describe "CApiObject" do
       it "returns nil if the instance variable has not been initialized" do
         @o.rb_ivar_get(@test, :@bar).should == nil
       end
+
+      it "returns nil if the instance variable has not been initialized and is not a valid Ruby name" do
+        @o.rb_ivar_get(@test, :bar).should == nil
+      end
+
+      it 'returns the instance variable when it is not a valid Ruby name' do
+        @o.rb_ivar_set(@test, :foo, 27)
+        @o.rb_ivar_get(@test, :foo).should == 27
+      end
     end
 
     describe "rb_ivar_set" do
@@ -820,6 +829,15 @@ describe "CApiObject" do
         @o.rb_ivar_set(@test, :@foo, 42).should == 42
         @test.instance_eval { @foo }.should == 42
       end
+
+      it "sets and returns the instance variable on an object" do
+        @o.rb_ivar_set(@test, :@foo, 42).should == 42
+        @test.instance_eval { @foo }.should == 42
+      end
+
+      it 'sets and returns the instance variable when it is not a valid Ruby name' do
+        @o.rb_ivar_set(@test, :foo, 27).should == 27
+      end
     end
 
     describe "rb_ivar_defined" do
@@ -830,6 +848,33 @@ describe "CApiObject" do
       it "returns false if the instance variable is not defined" do
         @o.rb_ivar_defined(@test, :@bar).should == false
       end
+
+      it "does not throw an error if the instance variable is not a valid Ruby name" do
+        @o.rb_ivar_defined(@test, :bar).should == false
+      end
+    end
+
+    # The `generic_iv_tbl` table and `*_generic_ivar` functions are for mutable
+    # objects which do not store ivars directly in MRI such as RString, because
+    # there is no member iv_index_tbl (ivar table) such as in RObject and RClass.
+
+    describe "rb_copy_generic_ivar for objects which do not store ivars directly" do
+      it "copies the instance variables from one object to another" do
+        original = "abc"
+        original.instance_variable_set(:@foo, :bar)
+        clone = "def"
+        @o.rb_copy_generic_ivar(clone, original)
+        clone.instance_variable_get(:@foo).should == :bar
+      end
+    end
+
+    describe "rb_free_generic_ivar for objects which do not store ivars directly" do
+      it "removes the instance variables from an object" do
+        o = "abc"
+        o.instance_variable_set(:@baz, :flibble)
+        @o.rb_free_generic_ivar(o)
+        o.instance_variables.should == []
+      end
     end
   end
 end
