@@ -4,22 +4,7 @@ require_relative '../../shared/string/times'
 
 load_extension('string')
 
-describe :rb_str_new2, shared: true do
-  it "returns a new string object calling strlen on the passed C string" do
-    # Hardcoded to pass const char * = "hello\0invisible"
-    @s.send(@method, "hello\0not used").should == "hello"
-  end
-
-  it "encodes the string with ASCII_8BIT" do
-    @s.send(@method, "hello").encoding.should == Encoding::ASCII_8BIT
-  end
-end
-
-describe "C-API String function" do
-  before :each do
-    @s = CApiStringSpecs.new
-  end
-
+class CApiStringSpecs
   class ValidTostrTest
     def to_str
       "ruby"
@@ -32,30 +17,98 @@ describe "C-API String function" do
     end
   end
 
-  describe "rb_str_set_len" do
-    before :each do
-      # Make a completely new copy of the string
-      # for every example (#dup doesn't cut it).
-      @str = "abcdefghij"[0..-1]
+  class ToSOrInspect
+    def to_s
+      'A string'
     end
 
-    it "reduces the size of the string" do
-      @s.rb_str_set_len(@str, 5).should == "abcde"
+    def inspect
+      'A different string'
     end
+  end
+end
 
-    it "inserts a NULL byte at the length" do
-      @s.rb_str_set_len(@str, 5).should == "abcde"
-      @s.rb_str_set_len(@str, 8).should == "abcde\x00gh"
-    end
+describe :rb_str_new2, shared: true do
+  it "returns a new string object calling strlen on the passed C string" do
+    # Hardcoded to pass const char * = "hello\0invisible"
+    @s.send(@method, "hello\0not used").should == "hello"
+  end
 
-    it "updates the string's attributes visible in C code" do
-      @s.rb_str_set_len_RSTRING_LEN(@str, 4).should == 4
+  it "encodes the string with ASCII_8BIT" do
+    @s.send(@method, "hello").encoding.should == Encoding::BINARY
+  end
+end
+
+describe "C-API String function" do
+  before :each do
+    @s = CApiStringSpecs.new
+  end
+
+  [Encoding::BINARY, Encoding::UTF_8].each do |enc|
+    describe "rb_str_set_len on a #{enc.name} String" do
+      before :each do
+        @str = "abcdefghij".force_encoding(enc)
+        # Make sure to unshare the string
+        @s.rb_str_modify(@str)
+      end
+
+      it "reduces the size of the string" do
+        @s.rb_str_set_len(@str, 5).should == "abcde"
+      end
+
+      it "inserts a NULL byte at the length" do
+        @s.rb_str_set_len(@str, 5).should == "abcde"
+        @s.rb_str_set_len(@str, 8).should == "abcde\x00gh"
+      end
+
+      it "updates the byte size" do
+        @s.rb_str_set_len(@str, 4)
+        @str.bytesize.should == 4
+        @str.should == "abcd"
+      end
+
+      it "invalidates the character size" do
+        @str.size.should == 10
+        @s.rb_str_set_len(@str, 4)
+        @str.size.should == 4
+        @str.should == "abcd"
+      end
+
+      it "invalidates the code range" do
+        @s.rb_str_set_len(@str, 4)
+        @str.ascii_only?.should == true
+      end
+
+      it "updates the string's attributes visible in C code" do
+        @s.rb_str_set_len_RSTRING_LEN(@str, 4).should == 4
+      end
+
+      it "can reveal characters written from C with RSTRING_PTR" do
+        @s.rb_str_set_len(@str, 1)
+        @str.should == "a"
+
+        @s.RSTRING_PTR_set(@str, 1, 'B'.ord)
+        @s.RSTRING_PTR_set(@str, 2, 'C'.ord)
+        @s.rb_str_set_len(@str, 3)
+
+        @str.bytesize.should == 3
+        @str.should == "aBC"
+      end
     end
   end
 
   describe "rb_str_buf_new" do
     it "returns the equivalent of an empty string" do
-      @s.rb_str_buf_new(10, nil).should == ""
+      buf = @s.rb_str_buf_new(10, nil)
+      buf.should == ""
+      buf.bytesize.should == 0
+      buf.size.should == 0
+      @s.RSTRING_LEN(buf).should == 0
+    end
+
+    it "returns a string with the given capacity" do
+      buf = @s.rb_str_buf_new(256, nil)
+      @s.rb_str_capacity(buf).should == 256
     end
 
     it "returns a string that can be appended to" do
@@ -83,6 +136,19 @@ describe "C-API String function" do
       str[0, 6].should == "abcd\x00f"
       @s.RSTRING_LEN(str).should == 8
     end
+
+    it "can be used as a general buffer and reveal characters with rb_str_set_len" do
+      str = @s.rb_str_buf_new(10, "abcdef")
+
+      @s.RSTRING_PTR_set(str, 0, 195)
+      @s.RSTRING_PTR_set(str, 1, 169)
+      @s.rb_str_set_len(str, 2)
+
+      str.force_encoding(Encoding::UTF_8)
+      str.bytesize.should == 2
+      str.size.should == 1
+      str.should == "é"
+    end
   end
 
   describe "rb_str_buf_new2" do
@@ -93,14 +159,27 @@ describe "C-API String function" do
   end
 
   describe "rb_str_new" do
+    it "creates a new String with BINARY Encoding" do
+      @s.rb_str_new("", 0).encoding.should == Encoding::BINARY
+    end
+
     it "returns a new string object from a char buffer of len characters" do
       @s.rb_str_new("hello", 3).should == "hel"
     end
 
+    it "returns a non-tainted string" do
+      @s.rb_str_new("hello", 5).tainted?.should == false
+    end
+
     it "returns an empty string if len is 0" do
       @s.rb_str_new("hello", 0).should == ""
     end
 
+    it "copy length bytes and does not stop at the first \\0 byte" do
+      @s.rb_str_new("he\x00llo", 6).should == "he\x00llo"
+      @s.rb_str_new_native("he\x00llo", 6).should == "he\x00llo"
+    end
+
     it "returns a string from an offset char buffer" do
       @s.rb_str_new_offset("hello", 1, 3).should == "ell"
     end
@@ -110,12 +189,6 @@ describe "C-API String function" do
     it_behaves_like :rb_str_new2, :rb_str_new2
   end
 
-  describe "rb_str_new" do
-    it "creates a new String with ASCII-8BIT Encoding" do
-      @s.rb_str_new("", 0).encoding.should == Encoding::ASCII_8BIT
-    end
-  end
-
   describe "rb_str_new_cstr" do
     it_behaves_like :rb_str_new2, :rb_str_new_cstr
   end
@@ -254,7 +327,7 @@ describe "C-API String function" do
     end
 
     it "raises a TypeError trying to append non-String-like object" do
-      lambda { @s.rb_str_append("Hello", 32323)}.should raise_error(TypeError)
+      -> { @s.rb_str_append("Hello", 32323)}.should raise_error(TypeError)
     end
 
     it "changes Encoding if a string is appended to an empty string" do
@@ -270,7 +343,7 @@ describe "C-API String function" do
   end
 
   describe "rb_str_times" do
-    it_behaves_like :string_times, :rb_str_times, ->(str, times) { @s.rb_str_times(str, times) }
+    it_behaves_like :string_times, :rb_str_times, -> str, times { @s.rb_str_times(str, times) }
   end
 
   describe "rb_str_buf_cat" do
@@ -341,7 +414,7 @@ describe "C-API String function" do
     end
 
     it "converts a C string to a Fixnum strictly if base is 0" do
-      lambda { @s.rb_cstr2inum("1234a", 0) }.should raise_error(ArgumentError)
+      -> { @s.rb_cstr2inum("1234a", 0) }.should raise_error(ArgumentError)
     end
   end
 
@@ -359,7 +432,7 @@ describe "C-API String function" do
     end
 
     it "converts a C string to a Fixnum strictly" do
-      lambda { @s.rb_cstr_to_inum("1234a", 10, true) }.should raise_error(ArgumentError)
+      -> { @s.rb_cstr_to_inum("1234a", 10, true) }.should raise_error(ArgumentError)
     end
   end
 
@@ -381,12 +454,12 @@ describe "C-API String function" do
   describe "rb_str_to_str" do
     it "calls #to_str to coerce the value to a String" do
       @s.rb_str_to_str("foo").should == "foo"
-      @s.rb_str_to_str(ValidTostrTest.new).should == "ruby"
+      @s.rb_str_to_str(CApiStringSpecs::ValidTostrTest.new).should == "ruby"
     end
 
     it "raises a TypeError if coercion fails" do
-      lambda { @s.rb_str_to_str(0) }.should raise_error(TypeError)
-      lambda { @s.rb_str_to_str(InvalidTostrTest.new) }.should raise_error(TypeError)
+      -> { @s.rb_str_to_str(0) }.should raise_error(TypeError)
+      -> { @s.rb_str_to_str(CApiStringSpecs::InvalidTostrTest.new) }.should raise_error(TypeError)
     end
   end
 
@@ -402,7 +475,7 @@ describe "C-API String function" do
 
     it "allows changing the characters in the string" do
       str = "abc"
-      @s.RSTRING_PTR_assign(str, 65)
+      @s.RSTRING_PTR_assign(str, 'A'.ord)
       str.should == "AAA"
     end
 
@@ -417,6 +490,13 @@ describe "C-API String function" do
       ret.should == str
     end
 
+    it "reflects changes from native memory and from String#setbyte in bounds" do
+      str = "abc"
+      from_rstring_ptr = @s.RSTRING_PTR_after_yield(str) { str.setbyte(1, 'B'.ord) }
+      from_rstring_ptr.should == "1B2"
+      str.should == "1B2"
+    end
+
     it "returns a pointer to the contents of encoded pointer-sized string" do
       s = "70パク".
         encode(Encoding::UTF_16LE).
@@ -469,7 +549,7 @@ describe "C-API String function" do
     it "does not call #to_s on non-String objects" do
       str = mock("fake")
       str.should_not_receive(:to_s)
-      lambda { @s.send(@method, str) }.should raise_error(TypeError)
+      -> { @s.send(@method, str) }.should raise_error(TypeError)
     end
   end
 
@@ -482,7 +562,7 @@ describe "C-API String function" do
       begin
         Thread.new {
           $SAFE = 1
-          lambda {
+          -> {
             @s.SafeStringValue("str".taint)
           }.should raise_error(SecurityError)
         }.join
@@ -494,10 +574,33 @@ describe "C-API String function" do
     it_behaves_like :string_value_macro, :SafeStringValue
   end
 
+  describe "rb_str_modify_expand" do
+    it "grows the capacity to bytesize + expand, not changing the bytesize" do
+      str = @s.rb_str_buf_new(256, "abcd")
+      @s.rb_str_capacity(str).should == 256
+
+      @s.rb_str_set_len(str, 3)
+      str.bytesize.should == 3
+      @s.RSTRING_LEN(str).should == 3
+      @s.rb_str_capacity(str).should == 256
+
+      @s.rb_str_modify_expand(str, 4)
+      str.bytesize.should == 3
+      @s.RSTRING_LEN(str).should == 3
+      @s.rb_str_capacity(str).should == 7
+
+      @s.rb_str_modify_expand(str, 1024)
+      str.bytesize.should == 3
+      @s.RSTRING_LEN(str).should == 3
+      @s.rb_str_capacity(str).should == 1027
+    end
+  end
+
   describe "rb_str_resize" do
     it "reduces the size of the string" do
       str = @s.rb_str_resize("test", 2)
       str.size.should == 2
+      str.bytesize.should == 2
       @s.RSTRING_LEN(str).should == 2
       str.should == "te"
     end
@@ -510,6 +613,7 @@ describe "C-API String function" do
       expected = "test".force_encoding("US-ASCII")
       str = @s.rb_str_resize(expected.dup, 12)
       str.size.should == 12
+      str.bytesize.should == 12
       @s.RSTRING_LEN(str).should == 12
       str[0, 4].should == expected
     end
@@ -572,10 +676,10 @@ describe :rb_external_str_new, shared: true do
     @s.send(@method, "abc").encoding.should == Encoding::UTF_8
   end
 
-  it "returns an ASCII-8BIT encoded string if any non-ascii bytes are present and default external is US-ASCII" do
+  it "returns a binary encoded string if any non-ascii bytes are present and default external is US-ASCII" do
     Encoding.default_external = "US-ASCII"
     x80 = [0x80].pack('C')
-    @s.send(@method, "#{x80}abc").encoding.should == Encoding::ASCII_8BIT
+    @s.send(@method, "#{x80}abc").encoding.should == Encoding::BINARY
   end
 
   it "returns a tainted String" do
@@ -631,10 +735,10 @@ describe "C-API String function" do
       s.encoding.should == Encoding::UTF_8
     end
 
-    it "returns an ASCII-8BIT encoded String if any non-ascii bytes are present and the specified encoding is US-ASCII" do
+    it "returns a binary encoded String if any non-ascii bytes are present and the specified encoding is US-ASCII" do
       x80 = [0x80].pack('C')
       s = @s.rb_external_str_new_with_enc("#{x80}abc", 4, Encoding::US_ASCII)
-      s.encoding.should == Encoding::ASCII_8BIT
+      s.encoding.should == Encoding::BINARY
     end
 
 
@@ -646,7 +750,7 @@ describe "C-API String function" do
 #         s = @s.rb_external_str_new_with_enc(a, a.bytesize, Encoding::UTF_8)
 #  -
 #  -      s.should == "\xA4\xA2\xA4\xEC".force_encoding("euc-jp")
-#  +      x = [0xA4, 0xA2, 0xA4, 0xEC].pack('C4')#.force_encoding('ascii-8bit')
+#  +      x = [0xA4, 0xA2, 0xA4, 0xEC].pack('C4')#.force_encoding('binary')
 #  +      s.should == x
 #         s.encoding.should equal(Encoding::EUC_JP)
 #     end
@@ -713,9 +817,9 @@ describe "C-API String function" do
         @s.rb_str_conv_enc(a, Encoding::UTF_8, Encoding::US_ASCII).should equal(a)
       end
 
-      it "returns the origin String if the destination encoding is ASCII-8BIT" do
-        a = "abc".force_encoding("ascii-8bit")
-        @s.rb_str_conv_enc(a, Encoding::US_ASCII, Encoding::ASCII_8BIT).should equal(a)
+      it "returns the origin String if the destination encoding is BINARY" do
+        a = "abc".force_encoding("binary")
+        @s.rb_str_conv_enc(a, Encoding::US_ASCII, Encoding::BINARY).should equal(a)
       end
     end
   end
@@ -753,10 +857,10 @@ describe "C-API String function" do
                                 Encoding::US_ASCII, 0, nil).should equal(a)
       end
 
-      it "returns the origin String if the destination encoding is ASCII-8BIT" do
-        a = "abc".force_encoding("ascii-8bit")
+      it "returns the origin String if the destination encoding is BINARY" do
+        a = "abc".force_encoding("binary")
         @s.rb_str_conv_enc_opts(a, Encoding::US_ASCII,
-                                Encoding::ASCII_8BIT, 0, nil).should equal(a)
+                                Encoding::BINARY, 0, nil).should equal(a)
       end
     end
   end
@@ -777,6 +881,27 @@ describe "C-API String function" do
     end
   end
 
+  describe "rb_str_export_to_enc" do
+    it "returns a copy of an ascii string converted to the new encoding" do
+      source = "A simple string".encode(Encoding::US_ASCII)
+      result = @s.rb_str_export_to_enc(source, Encoding::UTF_8)
+      result.should == source.encode(Encoding::UTF_8)
+      result.encoding.should == Encoding::UTF_8
+    end
+
+    it "returns the source string if it can not be converted" do
+      source = ["00ff"].pack("H*");
+      result = @s.rb_str_export_to_enc(source, Encoding::UTF_8)
+      result.should equal(source)
+    end
+
+    it "does not alter the source string if it can not be converted" do
+      source = ["00ff"].pack("H*");
+      result = @s.rb_str_export_to_enc(source, Encoding::UTF_8)
+      source.bytes.should == [0, 255]
+    end
+end
+
   describe "rb_sprintf" do
     it "replaces the parts like sprintf" do
       @s.rb_sprintf1("Awesome %s is replaced", "string").should == "Awesome string is replaced"
@@ -787,6 +912,26 @@ describe "C-API String function" do
       s = "Awesome %s is here with %s"
       @s.rb_sprintf2(s, "string", "content").should == "Awesome string is here with content"
     end
+
+    it "formats a string VALUE using to_s if sign not specified in format" do
+      s = 'Result: A string.'
+      @s.rb_sprintf3(CApiStringSpecs::ToSOrInspect.new).should == s
+    end
+
+    it "formats a string VALUE using inspect if sign specified in format" do
+      s = 'Result: A different string.'
+      @s.rb_sprintf4(CApiStringSpecs::ToSOrInspect.new).should == s
+    end
+
+    it "formats a TrueClass VALUE as `TrueClass` if sign not specified in format" do
+      s = 'Result: TrueClass.'
+      @s.rb_sprintf3(true.class).should == s
+    end
+
+    it "formats a TrueClass VALUE as 'true' if sign specified in format" do
+      s = 'Result: true.'
+      @s.rb_sprintf4(true.class).should == s
+    end
   end
 
   describe "rb_vsprintf" do
@@ -802,15 +947,54 @@ describe "C-API String function" do
     end
 
     it "tries to convert the passed argument to a string by calling #to_str first" do
-      @s.rb_String(ValidTostrTest.new).should == "ruby"
+      @s.rb_String(CApiStringSpecs::ValidTostrTest.new).should == "ruby"
     end
 
     it "raises a TypeError if #to_str does not return a string" do
-      lambda { @s.rb_String(InvalidTostrTest.new) }.should raise_error(TypeError)
+      -> { @s.rb_String(CApiStringSpecs::InvalidTostrTest.new) }.should raise_error(TypeError)
     end
 
     it "tries to convert the passed argument to a string by calling #to_s" do
       @s.rb_String({"bar" => "foo"}).should == '{"bar"=>"foo"}'
     end
   end
+
+  describe "rb_string_value_cstr" do
+    it "returns a non-null pointer for a simple string" do
+      @s.rb_string_value_cstr("Hello").should == true
+    end
+
+    it "returns a non-null pointer for a UTF-16 string" do
+      @s.rb_string_value_cstr("Hello".encode('UTF-16BE')).should == true
+    end
+
+    it "raises an error if a string contains a null" do
+      -> { @s.rb_string_value_cstr("Hello\0 with a null.") }.should raise_error(ArgumentError)
+    end
+
+    it "raises an error if a UTF-16 string contains a null" do
+      -> { @s.rb_string_value_cstr("Hello\0 with a null.".encode('UTF-16BE')) }.should raise_error(ArgumentError)
+    end
+
+  end
+
+  describe "rb_str_drop_bytes" do
+    it "drops N characters for an ASCII string" do
+      str = "12345678".encode("US-ASCII")
+      @s.rb_str_drop_bytes(str, 4)
+      str.should == "5678".encode("US-ASCII")
+    end
+
+    it "drop N/2 characters for a UTF-16 string" do
+      str = "12345678".encode("UTF-16LE")
+      @s.rb_str_drop_bytes(str, 4)
+      str.should == "345678".encode("UTF-16LE")
+    end
+
+    it "drop N/4 characters for a UTF-32 string" do
+      str = "12345678".encode("UTF-32LE")
+      @s.rb_str_drop_bytes(str, 4)
+      str.should == "2345678".encode("UTF-32LE")
+    end
+  end
 end
